#!/usr/bin/env python
#
# Copyright (c) James Cherti
# URL: https://github.com/jamescherti/elispcc
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <https://www.gnu.org/licenses/>.
#
"Recursively byte-compile and native-compile .el files."

import argparse
import os
import shutil
import subprocess
import sys

# Global variables
EMACS_BYTE_COMP_ENABLED = True
EMACS_NATIVE_COMP_ENABLED = True

# FYI: Native comp files will be saved to EMACS_D/eln-cache
EMACS_D = os.path.expanduser("~/.emacs.d")
EMACS_DEFAULT_ELN_CACHE_DIR = os.path.join(EMACS_D, "eln-cache")

EMACS_NATIVE_COMP_RESERVED_CPUS = "2"

LISP_CODE = """
(progn
  (setq byte-compile-warnings t)
  (setq byte-compile-verbose t)
  (let* ((user-emacs-directory (getenv "EMACS_D"))
         (default-directory (getenv "EMACS_BYTE_COMP_DIR"))
         (eln-path (getenv "EMACS_ELN_CACHE_DIR"))
         (reserved-cpus (string-to-number
                         (getenv "EMACS_NATIVE_COMP_RESERVED_CPUS")))
         (byte-comp-enabled
          (/= 0 (string-to-number (getenv "EMACS_BYTE_COMP_ENABLED"))))
         (native-comp-enabled
          (and (/= 0 (string-to-number (getenv "EMACS_NATIVE_COMP_ENABLED")))
               (featurep 'native-compile)
               (fboundp 'native-comp-available-p)
               (native-comp-available-p))))
    ;; SET ELN-CACHE DIR
    (when native-comp-enabled
      (cond ((fboundp 'startup-redirect-eln-cache) ; Emacs >= 28
             (startup-redirect-eln-cache eln-path))
            ((boundp 'native-comp-eln-load-path)
             (setcar native-comp-eln-load-path
                     (expand-file-name (convert-standard-filename eln-path)
                                       user-emacs-directory)))
            (t (error "Cannot change the eln-path directory"))))

    ;; LOAD DIRECTORIES
    (normal-top-level-add-subdirs-to-load-path)

    ;; SHOW MESSAGES
    (message "[INFO] Byte comp enabled: %s" native-comp-enabled)
    (message "[INFO] Native comp enabled: %s" byte-comp-enabled)
    (message "[INFO] Emacs user directory: %s\n" user-emacs-directory)

    ;; BYTE-COMP
    (when byte-comp-enabled
      (message "[TASK] Byte compile")
      (byte-recompile-directory default-directory 0))

    ;; NATIVE-COMP
    (when native-comp-enabled
      (message "\\n[TASK] Native compile")
      (setq native-comp-async-report-warnings-errors t)
      (setq native-comp-warning-on-missing-source t)
      (setq native-comp-async-jobs-number
            (max 1 (- (num-processors) reserved-cpus)))
      (setq native-comp-deferred-compilation nil)
      (setq package-native-compile nil)
      (native-compile-async default-directory 'recursively)
      (while comp-files-queue
        (sleep-for 0.1)))))
"""


class EmacsCompile:
    def __init__(self):
        self.args = None
        self._parse_args()

        self._flight_checks()
        self.update_os_environment()

        if not self.args.directories:
            print("Nothing to do.")
            sys.exit(1)

        for directory in self.args.directories:
            print(f"[INFO] Recursively compile the directory: {directory}")
            self.compile(directory)

    def _parse_args(self):
        """Parse command-line arguments."""
        usage = "%(prog)s [--option] [args]"
        parser = argparse.ArgumentParser(description=__doc__.splitlines()[0],
                                         usage=usage)
        parser.add_argument(
            "directories", metavar="N", type=str, nargs="*",
            help=("The directories to be scanned recursively by Emacs "
                  "to locate the '.el' files for compilation."),
        )

        parser.add_argument(
            "-c",
            "--eln-cache",
            default=EMACS_DEFAULT_ELN_CACHE_DIR,
            required=False,
            help=("The eln-cache directory where Emacs stores the "
                  "compiled native compiled code. Default: "
                  f"{EMACS_DEFAULT_ELN_CACHE_DIR}"),
        )
        self.args = parser.parse_args()

    def _flight_checks(self):
        if not EMACS_BYTE_COMP_ENABLED and not EMACS_NATIVE_COMP_ENABLED:
            print("Error: Neither byte compilation nor native compilation "
                  "is enabled.", file=sys.stderr)
            sys.exit(1)

        if not shutil.which("emacs"):
            print("Error: 'emacs' command not found or not executable.",
                  file=sys.stderr)
            sys.exit(1)

        if not os.path.isdir(self.args.eln_cache):
            try:
                os.mkdir(self.args.eln_cache)
            except OSError as err:
                print(f"Error: {err}")
                sys.exit(1)

        if not os.path.isdir(EMACS_D):
            print(f"Error: The directory does not exist: {EMACS_D}")
            sys.exit(1)

        for directory in self.args.directories:
            if not os.path.isdir(directory):
                print(f"Error: The directory does not exist: {directory}")
                sys.exit(1)

    def update_os_environment(self):
        os.environ["EMACS_D"] = EMACS_D
        os.environ["EMACS_NATIVE_COMP_RESERVED_CPUS"] = \
            str(EMACS_NATIVE_COMP_RESERVED_CPUS)
        os.environ["EMACS_NATIVE_COMP_ENABLED"] = \
            '1' if EMACS_NATIVE_COMP_ENABLED else '0'
        os.environ["EMACS_BYTE_COMP_ENABLED"] = \
            '1' if EMACS_BYTE_COMP_ENABLED else '0'
        os.environ["EMACS_ELN_CACHE_DIR"] = self.args.eln_cache

    @staticmethod
    def compile(directory: os.PathLike):
        env = os.environ.copy()
        env["EMACS_BYTE_COMP_DIR"] = directory

        emacs_cmd = ["emacs", "--batch", "--eval", LISP_CODE]
        try:
            subprocess.check_call(emacs_cmd, env=env,
                                  stderr=subprocess.STDOUT)
            print("Success.")
        except subprocess.CalledProcessError as err:
            print(f"Error: {err}")
            sys.exit(1)


def command_line_interface():
    """The command-line interface."""
    EmacsCompile()
